#!/Library/Frameworks/Python.framework/Versions/2.7/bin/python
#
# Last updated 16 Jan 2018
#
#============================================================================================
#
# Edit the "main" program (at the end of this file) to specify the name of a cluster
# and the year for which data is to be downloaded.
#
# Creates files:
#           
#
# Defines:
#      processCoincidences(cluster,year,month,csv3a,csvm)
#============================================================================================

import re           # Python standard library     
import math         # Python standard library
import datetime     # Python standard date and time utilities

import tables       # HDF/Saphire
import utilities
import solveSky

from solveSky  import celestialCoords3Station,celestialCoordsMultiStation,coincCoords, haToRa
from solveSky  import celestialCoordsMultiPerm,celestialCoordsMultiStation
from utilities import degToRad, radToDeg, mtonlgt,intraClusterDistance,raDecToGalactic,hammerAitoff,GPSTimeToDateAndTime,combinationDictionary


from sapphire                 import time_util,esd, download_data, CoincidencesESD, transformations
from sapphire.transformations import clock
from sapphire.api             import API, Network, Station



debug    = 1
testMode = True
maxrows  = 2000000


def processCoincidences(csv3,csvm,csvc,csvByMonth,csv3Extended,data,cluster,stationNumList,year,month):
    """ Process a coincidence HDF file for a specified year and month 
        (The year+month identifies the HDF file on disk by means of a standard name encoding: coincidences_<yyyy>_<m>.h5).
        Input:
            cluster:    (string)    cluster name (e.g. "Amsterdam")
            year:       integer     year (e.g. 2017)
            month:      integer     month in year, counting from 1 to 12.
            csv3        filehandle  the CSV file to which accumulated 3-station data will be written
            csvm        filehandle  the CSV file to which accumulated multi-station data will be written.
        Output:
            Produce an CSV output files containing altitude/azimuth and ra/declination coordinates.
            1) A CSV file containing 3-station coincidences for the current month: coincidences3<cluster>_<year>_<month>.csv
            2) A CSV file accumulating 3-station coincidences for all months processed in the current program run: coincidences3<cluster>.csv
            3) A CSV file accumulating multi-station coincidences for all months processed in the current program run: coincidencesM<cluster>.csv
    """
    
    global testMode
    testMode = False

    s_index = data.root.coincidences.s_index        # An array, each element of which is a station-id, in the form of a character string containing the 
                                                    # station number as its last part, for example: /hisparc/cluster_birmingham/station_14003.
                                                    # This will need to be transformed to a mapping from station number --> station name.
                                                    # (E.g   station_to_num["/hisparc/cluster_birmingham/station_14003"] == 14003.)
                                                    # The station number is used as an index.

    c_index = data.root.coincidences.c_index        # An array, each element of which corresponds one-to-one in sequence to elements in ctable[].
                                                    # Each array element of c_index is itself an array which each element relates to a station in the coincidence:
                                                    #   - each station element is a tuple: (seq_id, event_no)
                                                    #     where the seq_id indexes into s_index, to identify a station-id (a character string).
                                                    #     and event_no is the unique event identifier for the shower trace at that station.

    ctable  = data.root.coincidences.coincidences   # An array of coincidences.
                                                    # Each coincidence (ctable[i]) is a dictionary with the following entries:
                                                    #   ctable[i]['id']             Unique Id of the coincidence.
                                                    #   ctable[i]['N']              number of stations in the coincidence.
                                                    #   ctable[i]['timestamp']      unix format timestamp.
                                                    #   ctable[i]['nanosecs']       nanoseconds after the unix timestamp.
                                                    #   ctable[i]['ext_timestamp']  extended timestamp.


    # Create a dictionary relating station IDs (as used in the Coincidences data structure)
    # to station numbers (as used to access station information via the Station object).

    ( station_to_num, station_to_cluster ) = stationNumToIDdict(s_index)       
    
    scount3 = 0
    scountm = 0
    
    (combToID,idToComb)= combinationDictionary(stationNumList)
    
    if(debug >= 1):    
        print "Number of coincidences to process for month: ",month,"=",ctable.nrows
        if(maxrows > 0):
            print "Processing limited to ",maxrows," rows."

    for i in range(min(maxrows+1,ctable.nrows)):                   # Iterate over coincidences in the table.

        coincidence = ctable[i]
        
        coinc_multiplicity   = ctable[i]['N']   # The number of stations registering the coincidence 

        c_timestamp = coincidence['timestamp']        

        # Extract data and construct arrays to hold data describing a coincidence
        (coinc_stat_num,coinc_stat_id,coinc_gps,coinc_event,
                     coinc_lat,coinc_lng,
                     coinc_ecef_x, coinc_ecef_y, coinc_ecef_z,
                     coinc_delta,
                     coinc_ecef_dx,coinc_ecef_dy,coinc_ecef_dz,
                     coinc_enu_dx, coinc_enu_dy, coinc_enu_dz,
                     ecef_bac
                     ) = coincCoords(data, coincidence, station_to_num, c_index[i], s_index)
                     
        sortedStatNum = tuple(sorted(coinc_stat_num))
        
        try:
            config_id = combToID[sortedStatNum]
        except KeyError:
            print "Key Error: station list is ",sortedStatNum
            coinc_multiplicity = -1
            config_id = -1
        
        
        if(debug>=4):
            print "Coincidence %d"%i
            print coinc_stat_num
            print coinc_stat_id
            print coinc_delta
            print coinc_gps
            print coinc_event[0]
            print coinc_event[1]
            print coinc_event[2]

        cluster = station_to_cluster[coinc_stat_num[0]]
        
        timeString = GPSTimeToDateAndTime(coincidence['timestamp'])
        uts_timestamp = clock.gps_to_utc( coincidence['timestamp'] )
        
        (maxDistNanoLghtSecs,minDistNanoLgthSecs,meanDistNanoLghtSecs) = intraClusterDistance(coinc_ecef_dx,coinc_ecef_dy,coinc_ecef_dz)
        maxDist = maxDistNanoLghtSecs/mtonlgt
        minDist = minDistNanoLgthSecs/mtonlgt
        
        if( debug >= 2 ):
            print timeString,          
            print "  Multiplicity: ",coinc_multiplicity,            
            print "  ID = %4s "%(coincidence['id']),
            print "  Timestamp = %10d secs"%(c_timestamp),
            print " +%9d nanosecs"%(coincidence['nanoseconds']),
            print "  Maximum distance=%8.3f"%(maxDist),
            print "  Minimum distance=%8.3f"%(minDist),
            print "  Configuration ID=%6d"%(config_id),
            
        if coinc_multiplicity == 3:             # Need to use direct quadratic solution for exactly 3 stations.
            scount3 = scount3+1
            
            # Solve for the shower direction.
            (ha,dec,alt,azm,rms_dt_error,bac,ErrorMessage) = celestialCoords3Station( coinc_delta, coinc_ecef_dx, coinc_ecef_dy, coinc_ecef_dz,coinc_lat[0],coinc_lng[0])

            (ra,gmst)   = haToRa( ha, c_timestamp )                             # Convert solution 1 to Right Ascension
            (l,b)       = raDecToGalactic(ra,dec)                               # Convert to galactic coordinates.
            (xgal,ygal) = hammerAitoff(l,b)                                    # HammerAitoff equi-area projection of galactic coords.
            
            if( ErrorMessage == "" and bac*radToDeg > 20.0 and bac*radToDeg < 160.0):               # Distrust the results if the included angle between detector arms is too small or too big.
                eFlag = False
                writeCSVLine(csv3,      year,month,cluster,coincidence['id'],coinc_multiplicity,coincidence['timestamp'],uts_timestamp,gmst,maxDist,minDist,config_id,ra,ha,dec,alt,azm,l,b,xgal,ygal,rms_dt_error)
                
                writeCSV3ExtendedLine(csv3Extended,year,month,cluster,coincidence['id'], coinc_multiplicity, coincidence['timestamp'],uts_timestamp,
                                    gmst,maxDist,minDist,config_id,
                                    ra,ha,dec,alt,azm,l,b,xgal,ygal,
                                    coinc_stat_num[0],coinc_stat_num[1],coinc_stat_num[2],
                                    coinc_delta[1],coinc_delta[2],
                                    coinc_ecef_dx[1],coinc_ecef_dy[1],coinc_ecef_dz[1],
                                    coinc_ecef_dx[2],coinc_ecef_dy[2],coinc_ecef_dz[2],
                                    ecef_bac,ErrorMessage)
                
                
        
        if coinc_multiplicity >= 3:                 # Perform least squares solution.
            scountm = scountm+1
            eFlag = False
            
            (ha,dec,alt,azm,rms_dt_error)  = celestialCoordsMultiStation( coinc_delta, coinc_ecef_dx, coinc_ecef_dy, coinc_ecef_dz,  coinc_lat[0],coinc_lng[0] )
            (ra,gmst)  = haToRa( ha, c_timestamp )                             # Convert solution 1 t3o Right Ascension
            (l,b)      = raDecToGalactic(ra,dec)                               # Convert to galactic coordinates.
            (xgal,ygal)= hammerAitoff(l,b)                                     # HammerAitoff equi-area projection of galactic coords.
            
            
            # Calculate average of all 3-station combinations from the coincidence
            #(hac,decc,altc,azmc,rms_error) = celestialCoordsMultiPerm(coinc_delta, coinc_ecef_dx, coinc_ecef_dy, coinc_ecef_dz , coinc_lat[0],coinc_lng[0])
            #(rac,gmst) = haToRa( hac, c_timestamp )                             # Convert solution 1 to Right Ascension
            
            if(coinc_multiplicity > 3):                                          # File containinly only solutions for coincidences with more than 3 stations contibuting
                writeCSVLine(csvm,year,month,cluster,coincidence['id'],coinc_multiplicity,coincidence['timestamp'],uts_timestamp,gmst,maxDist,minDist,config_id,ra,ha,dec,alt,azm,l,b,xgal,ygal,rms_dt_error)
            
            writeCSVLine(csvc,      year,month,cluster,coincidence['id'],coinc_multiplicity,coincidence['timestamp'],uts_timestamp,gmst,maxDist,minDist,config_id,ra,ha,dec,alt,azm,l,b,xgal,ygal,rms_dt_error)   # Registers all coincidences
            writeCSVLine(csvByMonth,year,month,cluster,coincidence['id'],coinc_multiplicity,coincidence['timestamp'],uts_timestamp,gmst,maxDist,minDist,config_id,ra,ha,dec,alt,azm,l,b,xgal,ygal,rms_dt_error)   # Registers coincidences by month.     
            
            if( debug >= 2):
                print('ha=%6.1f, : ra=%7.1f dec=%6.1f : alt=%6.1f azm=%6.1f : l=%6.1f b=%6.1f : xgal=%6.3f ygal=%6.3f\n'%(radToDeg*ha,radToDeg*ra,radToDeg*dec,radToDeg*alt,radToDeg*azm,radToDeg*l,radToDeg*b,xgal,ygal)),
            
        #if( coinc_multiplicity > 3):
            #print(': hac=%6.1f : rac=%7.1f decc=%6.1f : altc=%6.1f,azmc%6.1f,'    %(radToDeg*hac,radToDeg*rac,radToDeg*decc,radToDeg*altc,radToDeg*azmc))
        #else:
            #print(" ")

#============================================================================================================


def stationNumToIDdict(s_index):
    """ Given a coincidences data structure, build two dictionaries:
         - station numbers ===> station ids
         - station numbers ===> station_cluster.
        Return (station_to_num, station_to_cluster)
    
        Preconditions:
        s_index =     An array, each element of which is a station-id, 
                      in the form of a character string containing the station number as its last part, 
                      for example: /hisparc/cluster_birmingham/station_14003.
                      This will need to be transformed to a mapping from station number --> station name.
                      (E.g   station_to_num["/hisparc/cluster_birmingham/station_14003"] == 14003.)
                      and    station_to_cluster[14003] == "birmingham"
                      
    """
    
    # Create empty dictionaries.
    station_to_num     = {}
    station_to_cluster = {}

    for s in s_index:
        values            = re.split('_',s)                                     #   Split the string on underscores.
        station_num       = int(values[-1])                                     #   The last element of this array will be the station number.
        station_to_num[s] = station_num                                         #   So copy this to the station number dictionary.
        cluster_string    = values[-2]                                          #   This currently has the form "<cluster>/station"
        station_to_cluster[station_num] = re.sub('/station','',cluster_string)  #   This maps station number to cluster name.

        if(debug >=3):
            print "Station: %-45s"%s,"==>  index= %5d"%station_to_num[s],"; cluster=",station_to_cluster[station_num]
            
    return ( station_to_num, station_to_cluster )            

#===================================================================================================================


def printStationCoordinates(stationNumbers):
    """Given a list of station numbers (integers), print station GPS coordinates.
    
       (This routine depends on the import of the saphire.API.Station subroutine.)
    """
    
    for stationNum in stationNumbers:
        station  = Station(stationNum)
        location = station.gps_location()
        print stationNum, station.country(), station.cluster(), station.n_detectors(),location['latitude'],location['longitude'], location['altitude']

#===================================================================================================================

def writeCSV3ExtendedHeading(csv3):
    """ Write the heading on the 3-station CSV table containing extended data."""
    csv3.write('year,month,cluster,id,multiplicity,')
    csv3.write('gps_timestamp,uts_timstamp,')
    csv3.write('GMST(Hours),GMST(deg),')
    csv3.write('maxDist(m),minDist(m),config_id,')
    csv3.write('ha(deg),ra(deg),dec(deg),alt(deg),azm(deg),l(deg),b(deg),xgal,ygal,')
    csv3.write('s1,s2,s3,')
    csv3.write('delta_ab,delta_ac,')
    csv3.write('dx1,dy1,dz1,')
    csv3.write('dx2,dy2,dz2,')
    csv3.write('ecef_bac,errorFlag')
    csv3.write('\n')
    
def writeCSV3ExtendedLine(csv3,year,month,cluster, coincid,coinc_multiplicity, timestamp,uts_timestamp,
                     gmst,maxDist,minDist,config_id, ra,ha,dec,alt,azm,l,b,xgal,ygal,  stat0,stat1,stat2,  dt1,dt2,  dx1,dy1,dz1,  dx2,dy2,dz2,  
                     ecef_bac,F):
    """Write an extended data file for 3-station solutions."""
                     
    csv3.write('%d,%d,%s,%d,%d,'    %(year,month,cluster,coincid, coinc_multiplicity))
    csv3.write('%d,%d,'             %(timestamp,uts_timestamp))
    csv3.write('%f,%f,'             %(gmst, radToDeg*gmst*math.pi/12.0))
    csv3.write('%f,%f,'             %(maxDist,minDist))
    csv3.write('%d,'                %(config_id))
    csv3.write('%f,%f,%f,%f,%f,'    %(radToDeg*ha,radToDeg*ra,radToDeg*dec,radToDeg*alt,radToDeg*azm)) 
    csv3.write('%f,%f,%f,%f,'       %(radToDeg*l,radToDeg*b,xgal,ygal))
    csv3.write('%d,%d,%d,'          %(stat0,stat1,stat2))
    csv3.write('%f,%f,'             %(dt1,dt2))
    csv3.write('%f,%f,%f,'          %(dx1,dy1,dz1))
    csv3.write('%f,%f,%f,'          %(dx2,dy2,dz2))
    csv3.write('%f,%s'              %(ecef_bac,F))
    csv3.write('\n')
    
def writeCSVHeading(csv):
    """ Write the heading on the 3-station CSV table."""
    csv.write('year,month,cluster,id,multiplicity,gps_timestamp,uts_timstamp,')
    csv.write('GMST(Hours),GMST(deg),')
    csv.write('maxDist(m),minDist(m),config_id,')
    csv.write('ha(deg),ra(deg),dec(deg),alt(deg),azm(deg),l(deg),b(deg),xgal,ygal,')
    csv.write('dt_error')
    csv.write('\n')
    
def writeCSVLine(csv,year,month,cluster,coinc_id,coinc_multiplicity,timestamp,uts_timestamp,gmst,maxDist,minDist,config_id,ra,ha,dec,alt,azm,l,b,xgal,ygal,rms_dt_error):
    """ Write the heading on the 3-station CSV table."""
            
    csv.write('%d,%d,%s,%d,%d,%d,%d,' %(year,month,cluster,coinc_id, coinc_multiplicity, timestamp,uts_timestamp))
    csv.write('%f,%f,'             %(gmst, radToDeg*gmst*math.pi/12.0))
    csv.write('%f,%f,'             %(maxDist,minDist))
    csv.write('%d,'                %(config_id))
    csv.write('%f,%f,%f,%f,%f,'    %(radToDeg*ha,radToDeg*ra,radToDeg*dec,radToDeg*alt,radToDeg*azm))
    csv.write('%f,%f,'             %(radToDeg*l,radToDeg*b))               
    csv.write('%f,%f,'             %(xgal,ygal))               
    csv.write('%f'                 %(rms_dt_error))
    csv.write('\n')

#===================================================================================================================
    
def processCluster(csv3,csvm,csvc,cluster,stationNumList,year,fromMonth,toMonth):
    """Iterate over months for a specified cluster, year and range of months."""
    
    for month in range(fromMonth,toMonth+1):             # for 2016 only 10 months of data as yet
        
        fileNameH5 = '%s/coincidence%s_%d_%d.h5'%(cluster,cluster,year,month)             # The INPUT file of coincidence data.
        data       = tables.open_file(fileNameH5,'r')                                     # The table containing the coincidence data.
        
        fileName3ExtCSV    = '%s/coincidence3_%s_%d_%d.csv'%(cluster,cluster,year,month)    # OUTPUT CSV file to hold 3 station coincidence outputs
        fileNameByMonthCSV = '%s/coincidence_%s_%d_%d.csv'%(cluster,cluster,year,month)     # OUTPUT CSV file to hold 3 station coincidence outputs
        
        if(debug >= 1):
            print "Process coincidences for cluster %s for year %d month=%2d"%(cluster,year,month),
            print "   Input from file: ",fileNameH5,
            
        
        csv3Extended = open(fileName3ExtCSV,'w')
        writeCSV3ExtendedHeading(csv3Extended)
        
        csvByMonth = open(fileNameByMonthCSV,'w')
        
        writeCSVHeading(csvByMonth)
        processCoincidences(csv3,csvm,csvc,csvByMonth,csv3Extended,data,cluster,stationNumList,year,month,)
        

def processFiles(cluster,stationNumList,year,fromMonth,toMonth):
    """Process data for a specified cluster, year and range of months."""
    
    fileNameCSV_3='%s/coincidence3_%s.csv'%(cluster,cluster)    # The file that will accumulate 3-station solutions
    fileNameCSV_M='%s/coincidenceM_%s.csv'%(cluster,cluster)    # The file that will accumulate >3-station solutions
    fileNameCSV_C='%s/coincidenceC_%s.csv'%(cluster,cluster)    # The file that will accumulate all solutions
    fileNameCSV_P='%s/coincidenceP_%s.csv'%(cluster,cluster)    # The file that will accumulate >3-station solutions evaluated by averaging combinations of 3 from multiplicity
    
    csv3 =open(fileNameCSV_3,'w')       # To hold 3 station solutions for all months
    csvm =open(fileNameCSV_M,'w')       # To hold multi-station solutions for all months
    csvc =open(fileNameCSV_C,'w')       # To hold combined solutions for all months
    
    writeCSVHeading(csv3)
    writeCSVHeading(csvm)
    writeCSVHeading(csvc)
    
    processCluster(csv3,csvm,csvc,cluster,stationNumList,year,fromMonth,toMonth)
                           
if __name__ == "__main__":
     stationNumList = [501,502,503, 504,505,506, 507,508,509, 510,511,521, 522]
     processFiles("Science_Park",stationNumList,year=2017,fromMonth=1,toMonth=12)





